"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const debug_1 = __importDefault(require("debug"));
const debug = debug_1.default('postqueue:queue');
/**
 * Represents a single or recurring job by ID.
 */
class Job {
    constructor(id, knex, tableOptions, pollMs, tx) {
        this.id = id;
        this.knex = knex;
        this.tableOptions = tableOptions;
        this.pollMs = pollMs;
        this.tx = tx;
    }
    /**
     * Waits for a job to complete (only applicable to non-recurring jobs)
     */
    done() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.tx) {
                throw new Error('Can\'t wait for job to finish inside processing function');
            }
            return new Promise((resolve, reject) => {
                const int = setInterval(() => __awaiter(this, void 0, void 0, function* () {
                    const res = yield this.knex.delete()
                        .from(this.tableOptions.resultTableName)
                        .where({
                        job_id: this.id
                    })
                        .returning('*');
                    if (res.length > 0) {
                        clearInterval(int);
                        yield this.knex.delete()
                            .from(this.tableOptions.tableName)
                            .where({ id: this.id });
                        resolve(res[0].result);
                    }
                }), this.pollMs);
            });
        });
    }
    /**
     * Removes job from queue
     */
    remove() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.tx) {
                yield this.tx.delete()
                    .where({ id: this.id })
                    .from(this.tableOptions.tableName);
            }
            else {
                yield this.knex.delete()
                    .where({ id: this.id })
                    .from(this.tableOptions.tableName);
            }
        });
    }
}
exports.Job = Job;
/**
 * This represents a single queue.
 */
class Queue {
    constructor(name, knex, options) {
        this.name = name;
        this.knex = knex;
        this.shouldProcess = true;
        const opts = Object.assign({
            tableName: 'postqueue',
            resultTableName: 'postqueue_results',
            pollInterval: 1000
        }, options);
        this.tableOptions = {
            tableName: opts.tableName,
            resultTableName: opts.resultTableName
        };
        this.pollInterval = opts.pollInterval;
    }
    /**
     * Adds a single or recurring job to the queue
     * @param data Job data
     * @param options Job options
     */
    add(data, options) {
        return __awaiter(this, void 0, void 0, function* () {
            const opts = Object.assign({
                everySecs: undefined,
                deleteOnAcknowledged: true
            }, options);
            if (opts.everySecs && !opts.deleteOnAcknowledged) {
                console.warn('deleteOnAcknowledged can only be false for single jobs. We always delete results for recurring jobs to prevent database clutter. Instead, you should provide results manually from inside the processing function.');
                opts.deleteOnAcknowledged = true;
            }
            const id = yield this.knex.insert({
                input_data: data,
                queue_name: this.name,
                every_secs: opts.everySecs,
                delete_on_acknowledged: opts.deleteOnAcknowledged,
                last_run: this.knex.raw('NOW()')
            })
                .returning('id')
                .into(this.tableOptions.tableName);
            return new Job(id[0], this.knex, this.tableOptions, this.pollInterval);
        });
    }
    /**
     * Gets a Job object from the queue.
     * @param id Job ID number
     */
    getJob(id) {
        return new Job(id, this.knex, this.tableOptions, this.pollInterval);
    }
    /**
     * Process runs the queue and processes jobs.
     * @param cb Callback to handle a job. Return value should be job result.
     */
    process(cb) {
        const knex = this.knex;
        this.shouldProcess = true;
        const toRun = () => {
            knex.transaction((trx) => __awaiter(this, void 0, void 0, function* () {
                const jobs = yield trx
                    .select(knex.raw('*'))
                    .forUpdate()
                    .skipLocked()
                    .from(this.tableOptions.tableName)
                    .limit(1)
                    .where('queue_name', this.name)
                    .andWhere(function () {
                    this.where('last_run', '<', knex.raw('NOW() - every_secs * interval \'1 seconds\''));
                    this.orWhereNull('every_secs');
                });
                debug(`got ${jobs.length} jobs`);
                if (jobs.length > 0) {
                    const job = jobs[0];
                    debug(`acquired lock for job ${job.id}`);
                    const jobPassed = new Job(job.id, knex, this.tableOptions, this.pollInterval, trx);
                    jobPassed.data = job.input_data;
                    const out = yield cb(jobPassed);
                    debug(`finished job ${job.id}, releasing lock`);
                    if (job.every_secs) {
                        debug('updating last run');
                        yield trx.update({
                            last_run: knex.raw(`
              CASE WHEN (last_run + every_secs * interval \'1 seconds\' > NOW() - every_secs * interval \'1 seconds\') THEN
                last_run + every_secs * interval \'1 seconds\'
                ELSE NOW()
              END`)
                        }).where({
                            id: job.id
                        }).into(this.tableOptions.tableName);
                    }
                    else {
                        if (!job.delete_on_acknowledged) { // delete_on_acknowledged == false implies they are expecting ack
                            debug('adding result');
                            yield trx.insert({
                                job_id: job.id,
                                result: out,
                                time_run: new Date()
                            }).into(this.tableOptions.resultTableName);
                        }
                        debug('deleting job');
                        yield trx.delete().where({
                            id: job.id
                        }).from(this.tableOptions.tableName);
                    }
                    yield trx.commit();
                    if (this.shouldProcess) {
                        setImmediate(toRun);
                    }
                }
                else {
                    if (this.shouldProcess) {
                        setTimeout(toRun, this.pollInterval);
                    }
                }
            }));
        };
        setImmediate(toRun);
    }
    shutdown() {
        this.shouldProcess = false;
    }
}
exports.Queue = Queue;
